/*
 * Autor: Piotr Buczkowski, 192902, N1-ISIII2
 *
 * Temat: boost::type_traits
 *
 * Opis dotyczy biblioteki w wersji 1.38
 *
 * Biblioteka "boost::type_traits" zawiera zbir specyficznych klas cech,
 * z ktrych kada obejmuje pojedyncz cech systemu C++.
 * Na przykad czy dany typ jest wskanikiem czy referencj?
 * Lub czy typ posiada konstruktor domylny albo kwalifikator "const"?
 *
 * Klas "type_traits" uywamy ich w programowaniu generycznym
 * do okrelenia waciwoci danego typu i optymalizacji
 * odpowiednich dla danego przypadku.
 *
 * Biblioteka "type_traits" zawiera take zbir klas
 * dokonujcych specyficznych transformacji na danym typie.
 * Potrafi na przykad usun kwalifikator "const" lub "volatile".
 * Kada klasa dokonujca transformacji definiuje
 * wewntrzny skadnik "type" bdcy typem transformacji.
 *
 */

/*
 * Szablon "integral_constant" jest w hierarchii dziedziczenia
 * przodkiem klas "true_type" oraz "false_type"
 */
template <class T, T val>
struct integral_constant
{
	typedef integral_constant<T, val> type;
	typedef T value_type;
	static const T value = val;
};

/*
 * Klasy "type_traits" maj wspln cech:
 * kada dziedziczy po "true_type" jeli ma okrelon waciwo
 * lub po "false_type" w przeciwnym wypadku.
 * Tak wic:
 * "true_type" jest typem oznaczajcym, e klasa posiada dan cech
 * "false_type" jest typem oznaczajcym, e klasa danej cechy nie posiada
 */
typedef integral_constant<bool, true> true_type;
typedef integral_constant<bool, false> false_type;

/*
 * Rozwamy najprostsz klas w bibliotece: "is_void<T>".
 * Dziedziczy ona po "true_type" tylko jeli jako typ "T" podamy "void".
 * Mwimy, e klasa szablonowa "is_void" dostarcza penej specjalizacji
 * (full-specialization), gdy typ "T" to "void". Jest to do wana technika,
 * ale czasem potrzebujemy czego pomidzy rozwizaniem generycznym
 * a pen specjalizacj. To dlatego komitet standaryzacyjny zdefiniowa
 * czciow specjalizacj (partial template-class specialization).
 */
template <typename T>
struct is_void : public false_type{};

template <>
struct is_void<void> : public true_type{};


/*
 * Kolejny przykad to klasa "is_pointer<T>".
 * Gowna wersja obsuguje przypadki gdy "T" nie jest wskanikiem,
 * a czciowa specjalizacja w przeciwnym razie.
 */
template <typename T>
struct is_pointer : public false_type{};

template <typename T>
struct is_pointer<T*> : public true_type{};

/*
 * Generalnie obowizuje zasada, e aby napisa czciow specjalizacj
 * naley uprzednio napisa gwn wersj.
 */


/*
 * Przyjrzyjmy si przykadowi pokazujcemu sposb w jaki klasy cech mog by uyte.
 * Pokazuje on w jaki sposb klasy cech mog by uyte
 * do podjcia decyzji o optymalizacji w czasie kompilacji.
 * Rozwamy algorytm kopiujcy z biblioteki standardowej:
 */
template <typename Iter1, typename Iter2>
Iter2 copy(Iter1 first, Iter1 last, Iter2 out);
/*
 * W pewnych okolicznociach kopiowanie moe by przeprowadzone najefektywniej
 * z uyciem funkcji "memcpy", ale by tak uczyni spenione musz by nastpujce warunki:
 *		- operator przypisania nie zosta zdefiniowany przez programist
 *		- brak skadnikw bdcych referencjami
 *		- wszystkie klasy bazowe oraz skadniki klasy musz mie trywialny operator przypisania
 *
 * Biblioteka "type_traits" dostarcza klas "has_trivial_assignment" tak,
 * e "has_trivial_assignment<T>::type" ma warto "true" tylko wtedy,
 * gdy typ "T" ma trywialny operator przypisania.
 */
/*
 * Rozwamy nastpujcy przykad:
 */
namespace detail
{
	/*
	 * Szablon funkcji robicej woln, ale bezpieczn kopi.
	 */
	template<typename I1, typename I2, bool b>
	I2 do_copy(I1 first, I1 last, I2 out, const integral_constant<bool, b>&)
	{
	   while(first != last)
	   {
		  *out = *first;
		  ++out;
		  ++first;
	   }
	   return out;
	}

	/*
	 * W tej wersji funkcji kopiujcej wymagamy,
	 * aby oba iteratory wskazyway na ten sam typ
	 * oraz by ostatni parametr by typu "true_type".
	 */
	template<typename T>
	T* do_copy(const T* first, const T* last, T* out, const true_type&)
	{
	   memcpy(out, first, (last-first)*sizeof(T));
	   return out+(last-first);
	}
}

/*
 * W tej wersji funkcji kopiujcej posugujemy si funkcj "do_copy"
 * przekazujc jej informacj o typie wartoci wskazywanych
 * przy uyciu "has_trivial_assign<value_type>".
 * Spowoduje to wybranie wersji funkcji "do_copy" z "memcpy"
 * tylko tam gdzie tego chcemy
 * (czyli jedynie dla typw posiadajcych trywialny operator przypisania)
 * lub wersji wolniejszej lecz bezpiecznej w przeciwnym przypadku.
 */
template<typename I1, typename I2>
inline I2 copy(I1 first, I1 last, I2 out)
{
   /*
    * Kopiujemy przy uyciu "memcpy" jeeli "T" ma trywialny operator przypisania
	* oraz jeli argumentami s wskaniki.
    */
   typedef typename std::iterator_traits<I1>::value_type value_type;
   return detail::copy_imp(first, last, out, boost::has_trivial_assign<value_type>());
}


/*
 * W nastpnym przykadzie rozwamy prost par dwch typw:
 */
template <typename T1, typename T2> 
struct pair 
{
	T1 first;
	T2 second;

	/*
	 * Ta prosta para nie moe przechowywa referencji,
	 * gdy konstruktor musiaby przyjmowa wtedy referencje do referencji.
	 */
	pair(const T1& nfirst, const T2& nsecond) : first(nfirst), second(nsecond) {}
};
/*
 * Aby zaradzi powyszemu problemowi posuymy si klas "add_reference",
 * ktra dodaje referencj do typu nie-referencyjnego.
 */
#include <boost/type_traits/add_reference.hpp>
template <typename T1, typename T2> 
struct improved_pair
{
	T1 first;
	T2 second;

	improved_pair(boost::add_reference<const T1>::type nfirst,
		boost::add_reference<const T2>::type nsecond) :
			first(nfirst), second(nsecond) {}
};


/*
 * Powysze przykady demonstroway ide klas cech
 * oraz to w jaki sposb mog one zosta uyte.
 * Kompletnej listy dostpnych klas naley szuka w dokumentacji "boost".
 */
